import shlex
import subprocess
import sys
from pathlib import Path
from typing import Any, override

import attrs
import git
from loguru import logger

from liblaf import grapes
from liblaf.cherries import core
from liblaf.cherries.typing import PathLike


@attrs.define
class Git(core.Run):
    commit: bool = True
    inputs: list[Path] = attrs.field(factory=list)
    outputs: list[Path] = attrs.field(factory=list)
    repo: git.Repo = attrs.field(default=None)
    temporaries: list[Path] = attrs.field(factory=list)
    verify: bool = False

    @override
    @core.impl(after=("Dvc",))
    def end(self, *args, **kwargs) -> None:
        if self.commit and self.repo.is_dirty(untracked_files=True):
            try:
                self.repo.git.add(all=True)
                subprocess.run(["git", "status"], check=False)
                message: str = self._make_commit_message()
                self.repo.git.commit(message=message, no_verify=not self.verify)
            except git.GitCommandError as err:
                logger.exception(err)
        self.plugin_root.log_other("cherries.git.sha", self.repo.head.commit.hexsha)

    @override
    @core.impl
    def log_input(self, path: PathLike, *args, **kwargs) -> None:
        path: Path = Path(path)
        self.inputs.append(path.relative_to(self.project_dir))

    @override
    @core.impl
    def log_output(
        self,
        path: PathLike,
        name: PathLike | None = None,
        **kwargs,
    ) -> None:
        path: Path = Path(path)
        self.outputs.append(path.relative_to(self.project_dir))

    @override
    @core.impl
    def log_temporary(
        self, path: PathLike, name: PathLike | None = None, **kwargs
    ) -> None:
        path: Path = Path(path)
        self.temporaries.append(path.relative_to(self.project_dir))

    @override
    @core.impl
    def start(self, *args, **kwargs) -> None:
        self.repo = git.Repo(search_parent_directories=True)

    def _make_commit_message(self) -> str:
        name: str = self.exp_name
        message: str = f"chore(cherries): {name}\n\n"
        meta: dict[str, Any] = {}
        if url := self.url:
            meta["url"] = url
        meta["exp_dir"] = self.exp_dir.relative_to(self.project_dir)
        meta["cwd"] = Path.cwd().relative_to(self.project_dir)
        meta["cmd"] = shlex.join(sys.orig_argv)
        if params := self.params:
            meta["params"] = params
        if inputs := self.inputs:
            meta["inputs"] = inputs
        if outputs := self.outputs:
            meta["outputs"] = outputs
        if temporaries := self.temporaries:
            meta["temporaries"] = temporaries
        message += grapes.yaml.encode(meta).decode()
        return message



from ._impl import ImplInfo, collect_impls, get_impl_info, impl
from ._plugin import Plugin
from ._run import (
    Run,
    active_run,
    end,
    log_asset,
    log_input,
    log_metric,
    log_metrics,
    log_other,
    log_others,
    log_output,
    log_parameter,
    log_parameters,
    start,
)
from ._spec import SpecInfo, collect_specs, spec
from ._utils import (
    PluginCachedProperty,
    PluginProperty,
    plugin_cached_property,
    plugin_property,
)
from .typing import MethodName, PluginId

__all__ = [
    "ImplInfo",
    "MethodName",
    "Plugin",
    "PluginCachedProperty",
    "PluginId",
    "PluginProperty",
    "Run",
    "SpecInfo",
    "active_run",
    "collect_impls",
    "collect_specs",
    "end",
    "get_impl_info",
    "impl",
    "log_asset",
    "log_input",
    "log_metric",
    "log_metrics",
    "log_other",
    "log_others",
    "log_output",
    "log_parameter",
    "log_parameters",
    "plugin_cached_property",
    "plugin_property",
    "spec",
    "start",
]



import functools
import inspect
from collections.abc import Callable, Mapping, Sequence
from typing import Any, Protocol, overload

import attrs
import wrapt

from liblaf import grapes

from .typing import MethodName


class Plugin(Protocol):
    def delegate(
        self,
        method: MethodName,
        args: Sequence[Any],
        kwargs: Mapping[str, Any],
        *,
        first_result: bool = False,
    ) -> Any: ...


@attrs.define
class SpecInfo:
    delegate: bool = attrs.field(default=True)
    first_result: bool = attrs.field(default=False)


@overload
def spec[C: Callable](
    func: C, /, *, delegate: bool = True, first_result: bool = False
) -> C: ...
@overload
def spec[C: Callable](
    *, delegate: bool = True, first_result: bool = False
) -> Callable[[C], C]: ...
def spec[**P, T](
    func: Callable[P, T] | None = None,
    /,
    *,
    delegate: bool = True,
    first_result: bool = False,
) -> Any:
    if func is None:
        return functools.partial(spec, delegate=delegate, first_result=first_result)

    info = SpecInfo(delegate=delegate, first_result=first_result)

    @wrapt.decorator
    def wrapper(
        wrapped: Callable[P, T],
        instance: Plugin,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> T:
        __tracebackhide__ = True
        if info.delegate:
            return instance.delegate(
                wrapped.__name__, args, kwargs, first_result=info.first_result
            )
        return wrapped(*args, **kwargs)

    func: Any = wrapper(func)
    grapes.wrapt_setattr(func, "spec", info)
    return func


def collect_specs(cls: type[Plugin] | Plugin) -> dict[str, SpecInfo]:
    if isinstance(cls, type):
        cls = type(cls)
    return {
        name: grapes.wrapt_getattr(method, "spec")
        for name, method in inspect.getmembers(
            cls, lambda m: grapes.wrapt_getattr(m, "spec", None) is not None
        )
    }



from . import config, core, meta, path_utils, plugins
from ._entrypoint import end, run, start
from ._version import __version__, __version_tuple__, version, version_tuple
from .config import BaseConfig, input, output, temporary  # noqa: A004
from .core import (
    Plugin,
    Run,
    active_run,
    log_asset,
    log_input,
    log_metric,
    log_metrics,
    log_other,
    log_others,
    log_output,
    log_parameter,
    log_parameters,
)
from .path_utils import (
    as_os_path,
    as_path,
    as_posix,
    data,
    entrypoint,
    exp_dir,
    params,
    path,
    project_dir,
    src,
)

__all__ = [
    "BaseConfig",
    "Plugin",
    "Run",
    "__version__",
    "__version_tuple__",
    "active_run",
    "as_os_path",
    "as_path",
    "as_posix",
    "config",
    "core",
    "data",
    "end",
    "entrypoint",
    "exp_dir",
    "input",
    "log_asset",
    "log_input",
    "log_metric",
    "log_metrics",
    "log_other",
    "log_others",
    "log_output",
    "log_parameter",
    "log_parameters",
    "meta",
    "output",
    "params",
    "path",
    "path_utils",
    "plugins",
    "project_dir",
    "run",
    "src",
    "start",
    "temporary",
    "version",
    "version_tuple",
]



from collections.abc import Generator
from pathlib import Path
from typing import override

import attrs

from ._abc import AssetResolver


@attrs.define
class AssetResolverVtk(AssetResolver):
    SUFFIXES: set[str] = attrs.field(
        factory=lambda: {
            ".obj",
            ".ply",
            ".stl",
            ".vti",
            ".vtk",
            ".vtp",
            ".vtr",
            ".vts",
            ".vtu",
        }
    )

    @override
    def match(self, path: Path) -> bool:
        return path.suffix in self.SUFFIXES

    @override
    def resolve(self, path: Path) -> Generator[Path]:
        if (landmarks := path.with_suffix(".landmarks.json")).exists():
            yield landmarks



from ._git import (
    GitInfo,
    git_auto_commit,
    git_branch,
    git_commit_sha,
    git_commit_url,
    git_info,
)
from ._name import exp_name, project_name

__all__ = [
    "GitInfo",
    "exp_name",
    "git_auto_commit",
    "git_branch",
    "git_commit_sha",
    "git_commit_url",
    "git_info",
    "project_name",
]



import os
import subprocess
from typing import override

import attrs

from liblaf.cherries import core
from liblaf.cherries.typing import PathLike


@attrs.define
class Dvc(core.Run):
    @override
    @core.impl(before=("Comet",))
    def log_asset(self, path: PathLike, *args, **kwargs) -> None:
        if self.check_ignore(path):
            return
        self._dvc("add", "--quiet", path)

    @override
    @core.impl(before=("Comet",))
    def log_input(self, *args, **kwargs) -> None:
        self.log_asset(*args, **kwargs)

    @override
    @core.impl(before=("Comet",))
    def log_output(self, *args, **kwargs) -> None:
        self.log_asset(*args, **kwargs)

    def _dvc(self, *args: str | bytes | os.PathLike[str] | os.PathLike[bytes]) -> None:
        subprocess.run(["dvc", *args], check=True)

    def check_ignore(self, path: PathLike) -> bool:
        process: subprocess.CompletedProcess[bytes] = subprocess.run(
            ["dvc", "check-ignore", "--quiet", path], check=False
        )
        return process.returncode == 0



from collections.abc import Generator
from pathlib import Path

import attrs

from .resolvers import AssetResolver, AssetResolverSeries, AssetResolverVtk


@attrs.define
class AssetResolverRegistry:
    _registry: dict[str, AssetResolver] = attrs.field(factory=dict)

    def register(self, resolver: AssetResolver) -> None:
        self._registry[resolver.id] = resolver

    def resolve(self, path: Path) -> Generator[Path]:
        for resolver in self._registry.values():
            if not resolver.match(path):
                continue
            yield from resolver.resolve(path)


asset_resolver_registry: AssetResolverRegistry = AssetResolverRegistry()
asset_resolver_registry.register(AssetResolverSeries())
asset_resolver_registry.register(AssetResolverVtk())



from collections.abc import Mapping
from pathlib import Path
from typing import Any, override

import attrs
from loguru import logger

from liblaf import grapes
from liblaf.cherries import core


@attrs.define
class Logging(core.Run):
    @property
    def log_file(self) -> Path:
        return self.exp_dir / "logs" / self.entrypoint.with_suffix(".log").name

    @override
    @core.impl
    def start(self, *args, **kwargs) -> None:
        grapes.logging.init(link=False, file=self.log_file)

    @override
    @core.impl
    def log_metric(
        self,
        name: str,
        value: Any,
        /,
        step: int | None = None,
        epoch: int | None = None,
        **kwargs,
    ) -> None:
        __tracebackhide__ = True
        if step is None:
            logger.info("{name}: {value}", name=name, value=value)
        else:
            logger.info(
                "step: {step}, {name}: {value}", step=step, name=name, value=value
            )

    @override
    @core.impl
    def log_metrics(
        self,
        dic: Mapping[str, Any],
        /,
        prefix: str | None = None,
        step: int | None = None,
        epoch: int | None = None,
        **kwargs,
    ) -> None:
        __tracebackhide__ = True
        if step is None:
            logger.info("{dic}", dic=dic)
        else:
            logger.info("step: {step}, {dic}", step=step, dic=dic)

    @override
    @core.impl
    def end(self, *args, **kwargs) -> None:
        self.plugin_root.log_asset(self.log_file, "run.log")



import math
import operator
from collections.abc import Mapping, Sequence
from typing import Any, Self

import attrs
import cachetools
import networkx as nx
from loguru import logger

from ._impl import ImplInfo, collect_impls, get_impl_info
from .typing import MethodName, PluginId


@attrs.define
class Plugin:
    plugins: dict[PluginId, "Plugin"] = attrs.field(factory=dict, kw_only=True)

    _plugin_parent: Self | None = attrs.field(default=None, kw_only=True)
    _cache_sort_plugins: cachetools.Cache[MethodName, Sequence["Plugin"]] = attrs.field(
        factory=lambda: cachetools.Cache(math.inf), init=False
    )

    @property
    def plugin_id(self) -> str:
        return type(self).__name__

    @property
    def plugin_root(self) -> Self:
        if self._plugin_parent is None:
            return self
        return self._plugin_parent.plugin_root

    def delegate(
        self,
        method: MethodName,
        args: Sequence[Any] = (),
        kwargs: Mapping[str, Any] = {},
        *,
        first_result: bool = False,
    ) -> Any:
        __tracebackhide__ = True
        plugins: Sequence[Plugin] = self._plugins_sort(method)
        if not plugins:
            if first_result:
                return None
            return []
        results: list[Any] = []
        for plugin in plugins:
            try:
                result: Any = getattr(plugin, method)(*args, **kwargs)
            except BaseException as e:
                if isinstance(e, (KeyboardInterrupt, SystemExit)):
                    raise
                logger.exception("Plugin {}", plugin.plugin_id)
            else:
                if result is None:
                    continue
                if first_result:
                    return result
                results.append(result)
        return results

    def register(self, plugin: "Plugin") -> None:
        impls: dict[MethodName, ImplInfo] = collect_impls(plugin)
        for name in impls:
            self._cache_sort_plugins.pop(name, None)
        plugin._plugin_parent = self  # noqa: SLF001
        self.plugins[plugin.plugin_id] = plugin

    def _plugins_sort_cache_key(self, method: MethodName) -> MethodName:
        return method

    @cachetools.cachedmethod(
        operator.attrgetter("_cache_sort_plugins"), key=_plugins_sort_cache_key
    )
    def _plugins_sort(self, method: str) -> Sequence["Plugin"]:
        plugin_infos: dict[str, ImplInfo] = {
            plugin_id: info
            for plugin_id, plugin in self.plugins.items()
            if (info := get_impl_info(getattr(plugin, method, None))) is not None
        }

        def key_fn(node: str) -> int:
            return plugin_infos[node].priority

        graph: nx.DiGraph[str] = nx.DiGraph()
        for plugin_id, impl_info in plugin_infos.items():
            graph.add_node(plugin_id)
            for after in impl_info.after:
                if after in plugin_infos:
                    graph.add_edge(after, plugin_id)
            for before in impl_info.before:
                if before in plugin_infos:
                    graph.add_edge(plugin_id, before)
        return tuple(
            plugin
            for plugin_id in nx.lexicographical_topological_sort(graph, key=key_fn)
            if (plugin := self.plugins.get(plugin_id)) is not None
        )



import datetime
import inspect
import itertools
from collections.abc import Callable, Mapping, Sequence
from typing import Any

import pydantic

from liblaf.cherries import config as _config
from liblaf.cherries import core, profiles


def end(*args, **kwargs) -> None:
    core.active_run.log_other("cherries.end_time", datetime.datetime.now().astimezone())
    core.active_run.end(*args, **kwargs)


def run[T](main: Callable[..., T], *, profile: profiles.ProfileLike | None = None) -> T:
    run: core.Run = start(profile=profile)
    args: Sequence[Any]
    kwargs: Mapping[str, Any]
    args, kwargs = _make_args(main)
    configs: list[pydantic.BaseModel] = [
        arg
        for arg in itertools.chain(args, *kwargs.values())
        if isinstance(arg, pydantic.BaseModel)
    ]
    for config in configs:
        run.log_parameters(_config.model_dump_without_assets(config, mode="json"))
        for path in _config.get_inputs(config):
            run.log_input(path)
    try:
        return main(*args, **kwargs)
    finally:
        for config in configs:
            for path in _config.get_outputs(config):
                run.log_output(path)
            for path in _config.get_temporaries(config):
                run.log_temporary(path)
        run.end()


def start(
    profile: profiles.ProfileLike | None = None,
) -> core.Run:
    run: core.Run = profiles.factory(profile).init()
    run.start()
    run.log_other("cherries.entrypoint", run.entrypoint.relative_to(run.project_dir))
    run.log_other("cherries.exp_dir", run.exp_dir.relative_to(run.project_dir))
    run.log_other("cherries.start_time", run.start_time)
    return run


def _make_args(func: Callable) -> tuple[Sequence[Any], Mapping[str, Any]]:
    signature: inspect.Signature = inspect.signature(func)
    args: list[Any] = []
    kwargs: dict[str, Any] = {}
    for name, param in signature.parameters.items():
        match param.kind:
            case (
                inspect.Parameter.POSITIONAL_ONLY
                | inspect.Parameter.POSITIONAL_OR_KEYWORD
            ):
                args.append(_make_arg(param))
            case inspect.Parameter.KEYWORD_ONLY:
                kwargs[name] = _make_arg(param)
            case _:
                pass
    return args, kwargs


def _make_arg(param: inspect.Parameter) -> Any:
    if param.default is not inspect.Parameter.empty:
        return param.default
    if param.annotation is not inspect.Parameter.empty:
        return param.annotation()
    return None



import unittest.mock
from collections.abc import Mapping
from pathlib import Path
from typing import Any, override

import attrs
import comet_ml
import cytoolz as toolz
import dvc.api
import dvc.exceptions
import git
from loguru import logger

from liblaf import grapes
from liblaf.cherries import core, meta, path_utils
from liblaf.cherries.typing import PathLike


@attrs.frozen
class Asset:
    path: PathLike
    name: PathLike | None
    metadata: Mapping[str, Any] | None = None
    kwargs: Mapping[str, Any] = attrs.field(factory=dict)


@attrs.define
class Comet(core.Run):
    disabled: bool = attrs.field(default=False)
    enable_dvc: bool = False
    _assets_git: list[Asset] = attrs.field(factory=list)

    @override
    @core.impl(after=("Git", "Logging"))
    def end(self, *args, **kwargs) -> None:
        try:
            self._log_asset_git_end()
        except git.GitError as err:
            logger.exception(err)
        self.experiment.end()

    @override
    @core.impl
    def get_others(self) -> Mapping[str, Any]:
        return self.experiment.others

    @override
    @core.impl
    def get_params(self) -> Mapping[str, Any]:
        return self.experiment.params

    @override
    @core.impl
    def get_url(self) -> str:
        return self.experiment.url  # pyright: ignore[reportReturnType]

    @override
    @core.impl(after=("Dvc",))
    def log_asset(
        self,
        path: PathLike,
        name: PathLike | None = None,
        **kwargs,
    ) -> None:
        if self._log_asset_git(path, name, **kwargs):
            return
        if self.enable_dvc and self._log_asset_dvc(path, name, **kwargs):
            return
        name = path_utils.as_posix(name)
        self.experiment.log_asset(path, name, **kwargs)

    @override
    @core.impl(after=("Dvc",))
    def log_input(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None:
        if name is None:
            path = Path(path)
            name = f"inputs/{path.name}"
        metadata = toolz.assoc(metadata or {}, "type", "input")
        self.log_asset(path, name, metadata=metadata, **kwargs)

    @override
    @core.impl
    def log_metric(self, *args, **kwargs) -> None:
        return self.experiment.log_metric(*args, **kwargs)

    @override
    @core.impl
    def log_metrics(self, *args, **kwargs) -> None:
        return self.experiment.log_metrics(*args, **kwargs)

    @override
    @core.impl
    def log_other(self, *args, **kwargs) -> None:
        return self.experiment.log_other(*args, **kwargs)

    @override
    @core.impl
    def log_others(self, *args, **kwargs) -> None:
        return self.experiment.log_others(*args, **kwargs)

    @override
    @core.impl(after=("Dvc",))
    def log_output(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None:
        if name is None:
            path = Path(path)
            name = f"outputs/{path.name}"
        metadata = toolz.assoc(metadata or {}, "type", "output")
        self.log_asset(path, name, metadata=metadata, **kwargs)

    @override
    @core.impl
    def log_parameter(self, *args, **kwargs) -> None:
        return self.experiment.log_parameter(*args, **kwargs)

    @override
    @core.impl
    def log_parameters(self, *args, **kwargs) -> None:
        return self.experiment.log_parameters(*args, **kwargs)

    @override
    @core.impl(after=("Logging",))
    def start(self, *args, **kwargs) -> None:
        logger.disable("comet_ml")
        try:
            comet_ml.start(
                project_name=self.project_name,
                experiment_config=comet_ml.ExperimentConfig(
                    disabled=self.disabled, name=self.exp_name
                ),
            )
        except ValueError as err:
            logger.warning(err)
        logger.enable("comet_ml")

    @property
    def experiment(self) -> comet_ml.CometExperiment:
        return comet_ml.get_running_experiment() or unittest.mock.MagicMock()

    def _log_asset_dvc(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> bool:
        path = Path(path)
        name = path_utils.as_posix(name)
        try:
            # ? I don't know why, but `dvc.api.get_url` only works with this. Maybe a DVC bug?
            dvc_path: Path = path.absolute().relative_to(Path.cwd())
            uri: str = dvc.api.get_url(str(dvc_path))
        except dvc.exceptions.OutputNotFoundError:
            return False
        dvc_file: Path = path.with_name(path.name + ".dvc")
        dvc_meta: Mapping[str, Any] = grapes.yaml.load(dvc_file)
        metadata: dict[str, Mapping] = toolz.merge(metadata or {}, dvc_meta["outs"][0])
        self.experiment.log_remote_asset(uri, name, metadata=metadata, **kwargs)
        return True

    def _log_asset_git(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> bool:
        try:
            repo = git.Repo(search_parent_directories=True)
        except git.InvalidGitRepositoryError:
            return False
        try:
            if repo.ignored(path):
                return False
        except git.GitCommandError:
            # `path` may be outside repository
            return False
        self._assets_git.append(
            Asset(path=path, name=name, metadata=metadata, kwargs=kwargs)
        )
        return True

    def _log_asset_git_end(self) -> None:
        if len(self._assets_git) == 0:
            return
        repo = git.Repo(search_parent_directories=True)
        info: meta.GitInfo = meta.git_info()
        for asset in self._assets_git:
            uri: str
            match str(info.platform):
                case "github":
                    path: Path = Path(asset.path).absolute()
                    path_rel: str = path.relative_to(repo.working_tree_dir).as_posix()  # pyright: ignore[reportArgumentType]
                    sha: str = repo.head.commit.hexsha
                    uri = f"https://{info.host}/{info.owner}/{info.repo}/raw/{sha}/{path_rel}"
                case _:
                    uri = path_utils.as_posix(asset.path)
            self.experiment.log_remote_asset(
                uri,
                path_utils.as_posix(asset.name),
                metadata=dict(asset.metadata) if asset.metadata is not None else None,
                **asset.kwargs,
            )


def _get_api_key() -> str | None:
    config: comet_ml.config.Config = comet_ml.get_config()
    return comet_ml.get_api_key(None, config)



import sys
from pathlib import Path

import git
import git.exc
from loguru import logger

from liblaf.cherries import utils


@utils.cache
def entrypoint(*, absolute: bool = True) -> Path:
    path = Path(sys.argv[0])
    if absolute:
        return path.absolute()
    return path.relative_to(project_dir())


EXP_DIR_NAMES: set[str] = {"exp", "experiment", "experiments", "exps", "src"}


@utils.cache
def exp_dir(*, absolute: bool = True) -> Path:
    entrypoint_: Path = entrypoint(absolute=True)
    parent: Path = entrypoint_.parent
    exp_dir: Path
    exp_dir = parent.parent if parent.name in EXP_DIR_NAMES else parent
    if absolute:
        return exp_dir
    exp_dir = exp_dir.relative_to(project_dir())
    return exp_dir


@utils.cache
def project_dir() -> Path:
    exp_dir_: Path = exp_dir(absolute=True)
    try:
        repo = git.Repo(exp_dir_, search_parent_directories=True)
    except git.exc.InvalidGitRepositoryError:
        logger.warning("Not in a git repository, using current directory", once=True)
        return exp_dir_
    else:
        return Path(repo.working_dir)



from pathlib import Path

import git.exc

from liblaf.cherries import path_utils

from ._git import GitInfo, git_info


def project_name() -> str | None:
    try:
        info: GitInfo = git_info()
    except git.exc.InvalidGitRepositoryError:
        return None
    else:
        return info.repo


def exp_name() -> str:
    exp_dir: Path = path_utils.exp_dir(absolute=False)
    exp_name: str = path_utils.as_posix(exp_dir)
    exp_name = exp_name.removeprefix("exp")
    exp_name = exp_name.removeprefix("/")
    return exp_name



import enum
import os
from collections.abc import Callable, Generator, Iterable
from pathlib import Path
from typing import Any, Literal

import attrs
import pydantic

import liblaf.cherries.path_utils as pu
from liblaf import grapes
from liblaf.cherries.typing import PathLike

from ._registry import asset_resolver_registry

type Extra = (
    PathLike
    | Iterable[PathLike]
    | Callable[[PathLike], PathLike | Iterable[PathLike | None] | None]
    | None
)


class AssetKind(enum.StrEnum):
    INPUT = enum.auto()
    OUTPUT = enum.auto()
    TEMPORARY = enum.auto()


@attrs.define
class MetaAsset:
    kind: AssetKind
    extra: Extra = None

    def resolve(self, value: Path) -> Generator[Path]:
        if self.extra is None:
            yield from asset_resolver_registry.resolve(value)
            return
        extra: PathLike | Iterable[PathLike | None] | None = (
            self.extra(value) if callable(self.extra) else self.extra  # noqa: S610
        )
        for p in grapes.as_iterable(extra, base_type=(str, bytes, os.PathLike)):
            if p is None:
                continue
            yield Path(p)


def asset(path: PathLike, extra: Extra = None, *, kind: AssetKind, **kwargs) -> Path:
    field_info: pydantic.fields.FieldInfo = pydantic.Field(pu.data(path), **kwargs)  # pyright: ignore[reportAssignmentType]
    field_info.metadata.append(MetaAsset(kind=kind, extra=extra))
    return field_info  # pyright: ignore[reportReturnType]


def get_assets(cfg: pydantic.BaseModel, kind: AssetKind) -> Generator[Path]:
    for name, info in type(cfg).model_fields.items():
        value: Any = getattr(cfg, name)
        if isinstance(value, pydantic.BaseModel):
            yield from get_assets(value, kind)
        for meta in info.metadata:
            if isinstance(meta, MetaAsset) and meta.kind == kind:
                value: Path = Path(value)
                yield value
                yield from meta.resolve(value)


def get_inputs(cfg: pydantic.BaseModel) -> Generator[Path]:
    yield from get_assets(cfg, AssetKind.INPUT)


def get_outputs(cfg: pydantic.BaseModel) -> Generator[Path]:
    yield from get_assets(cfg, AssetKind.OUTPUT)


def get_temporaries(cfg: pydantic.BaseModel) -> Generator[Path]:
    yield from get_assets(cfg, AssetKind.TEMPORARY)


def input(path: PathLike, extra: Extra = None, **kwargs) -> Path:  # noqa: A001
    return asset(path, extra=extra, kind=AssetKind.INPUT, **kwargs)


def model_dump_without_assets(
    model: pydantic.BaseModel,
    *,
    mode: str | Literal["json", "python"] = "json",  # noqa: PYI051
    **kwargs,
) -> dict[str, Any]:
    data: dict[str, Any] = model.model_dump(mode=mode, **kwargs)
    for name, info in type(model).model_fields.items():
        value: Any = getattr(model, name)
        if isinstance(value, pydantic.BaseModel):
            value = model_dump_without_assets(value)
        for meta in info.metadata:
            if isinstance(meta, MetaAsset):
                del data[name]
                break
        else:
            data[name] = value
    return data


def output(path: PathLike, extra: Extra = None, **kwargs) -> Path:
    return asset(path, extra=extra, kind=AssetKind.OUTPUT, **kwargs)


def temporary(path: PathLike, extra: Extra = None, **kwargs) -> Path:
    return asset(path, extra=extra, kind=AssetKind.TEMPORARY, **kwargs)



import abc
from collections.abc import Iterable
from pathlib import Path


class AssetResolver:
    @property
    def id(self) -> str:
        return type(self).__name__

    @abc.abstractmethod
    def match(self, path: Path) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def resolve(self, path: Path) -> Iterable[Path]:
        raise NotImplementedError



import shutil
from pathlib import Path
from typing import override

import attrs

from liblaf.cherries import core
from liblaf.cherries.typing import PathLike


@attrs.define
class Local(core.Run):
    folder: Path = attrs.field(default=None)

    @override
    @core.impl
    def log_asset(
        self,
        path: PathLike,
        name: PathLike | None = None,
        **kwargs,
    ) -> None:
        path = Path(path)
        if name is None:
            path_absolute: Path = path.resolve()
            if path_absolute.is_relative_to(self.exp_dir):
                name = path_absolute.relative_to(self.exp_dir)
            else:
                name = path
        target: Path = self.folder / name
        self._copy(path, target)

    @override
    @core.impl
    def log_input(
        self,
        path: PathLike,
        name: PathLike | None = None,
        **kwargs,
    ) -> None:
        if name is None:
            name = Path(path).name
        name = f"inputs/{name}"
        self.log_asset(path, name, **kwargs)

    @override
    @core.impl
    def log_output(
        self,
        path: PathLike,
        name: PathLike | None = None,
        **kwargs,
    ) -> None:
        if name is None:
            name = Path(path).name
        name = f"outputs/{name}"
        self.log_asset(path, name, **kwargs)

    @override
    @core.impl
    def log_temporary(
        self,
        path: PathLike,
        name: PathLike | None = None,
        **kwargs,
    ) -> None:
        if name is None:
            name = Path(path).name
        name = f"temp/{name}"
        self.log_asset(path, name, **kwargs)

    @override
    @core.impl
    def start(self, *args, **kwargs) -> None:
        cherries_dir: Path = self.exp_dir / ".cherries"
        cherries_dir.mkdir(parents=True, exist_ok=True)
        (cherries_dir / ".gitignore").write_text("*\n")
        self.folder = (
            cherries_dir
            / self.entrypoint.stem
            / self.start_time.strftime("%Y-%m-%dT%H%M%S")
        )
        entrypoint: Path = self.entrypoint
        self.log_asset(entrypoint, f"src/{entrypoint.name}")

    def _copy(self, source: PathLike, target: PathLike) -> None:
        source = Path(source)
        target = Path(target)
        target.parent.mkdir(parents=True, exist_ok=True)
        if source.is_dir():
            shutil.copytree(source, target)
        else:
            shutil.copy2(source, target)



import datetime
from collections.abc import Mapping
from pathlib import Path
from typing import Any

import attrs
from environs import env

from liblaf.cherries import path_utils
from liblaf.cherries.typing import PathLike

from ._plugin import Plugin
from ._spec import spec
from ._utils import plugin_cached_property, plugin_property


@attrs.define
class Run(Plugin):
    """.

    References:
        1. [Experiment - Comet Docs](https://www.comet.com/docs/v2/api-and-sdk/python-sdk/reference/Experiment/)
        2. [Logger | ClearML](https://clear.ml/docs/latest/docs/references/sdk/logger)
        3. [MLflow Tracking APIs | MLflow](https://www.mlflow.org/docs/latest/ml/tracking/tracking-api/)
    """

    @plugin_cached_property
    def data_dir(self) -> Path:
        return path_utils.data()

    @plugin_cached_property
    def entrypoint(self) -> Path:
        return path_utils.entrypoint()

    @plugin_cached_property
    def exp_dir(self) -> Path:
        return path_utils.exp_dir()

    @plugin_cached_property
    def exp_name(self) -> str:
        return (
            self.entrypoint.relative_to(self.project_dir)
            .as_posix()
            .removeprefix("exp/")
        )

    @plugin_property
    def params(self) -> Mapping[str, Any]:
        return self.plugin_root.get_params()

    @plugin_cached_property
    def project_name(self) -> str | None:
        return self.project_dir.name

    @plugin_cached_property
    def project_dir(self) -> Path:
        return path_utils.project_dir()

    @plugin_cached_property
    def start_time(self) -> datetime.datetime:
        return datetime.datetime.now().astimezone()

    @plugin_property
    def url(self) -> str:
        return self.plugin_root.get_url()

    @spec
    def end(self, *args, **kwargs) -> None: ...

    @spec(first_result=True)
    def get_others(self) -> Mapping[str, Any]: ...

    @spec(first_result=True)
    def get_params(self) -> Mapping[str, Any]: ...

    @spec(first_result=True)
    def get_url(self) -> str: ...

    @spec
    def log_asset(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None: ...

    @spec
    def log_input(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None: ...

    @spec
    def log_metric(
        self,
        name: str,
        value: Any,
        /,
        step: int | None = None,
        epoch: int | None = None,
        **kwargs,
    ) -> None: ...

    @spec
    def log_metrics(
        self,
        dic: Mapping[str, Any],
        /,
        prefix: str | None = None,
        step: int | None = None,
        epoch: int | None = None,
        **kwargs,
    ) -> None: ...

    @spec
    def log_other(self, key: Any, value: Any, /, **kwargs) -> None: ...

    @spec
    def log_others(self, dictionary: Mapping[Any, Any], /, **kwargs) -> None: ...

    @spec
    def log_output(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None: ...

    @spec
    def log_parameter(
        self, name: Any, value: Any, /, step: int | None = None, **kwargs
    ) -> None: ...

    @spec
    def log_parameters(
        self,
        parameters: Mapping[Any, Any],
        /,
        prefix: str | None = None,
        step: int | None = None,
        **kwargs,
    ) -> None: ...

    @spec
    def log_temporary(
        self,
        path: PathLike,
        name: PathLike | None = None,
        *,
        metadata: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None: ...

    @spec(delegate=False)
    def start(self, *args, **kwargs) -> None:
        env.read_env(self.entrypoint.parent / ".env")
        return self.delegate("start", args, kwargs)


active_run: Run = Run()
end = active_run.end
log_asset = active_run.log_asset
log_input = active_run.log_input
log_metric = active_run.log_metric
log_metrics = active_run.log_metrics
log_other = active_run.log_other
log_others = active_run.log_others
log_output = active_run.log_output
log_parameter = active_run.log_parameter
log_parameters = active_run.log_parameters
start = active_run.start



from . import resolvers
from ._meta import (
    AssetKind,
    Extra,
    MetaAsset,
    asset,
    get_assets,
    get_inputs,
    get_outputs,
    get_temporaries,
    input,  # noqa: A004
    model_dump_without_assets,
    output,
    temporary,
)
from ._registry import AssetResolverRegistry, asset_resolver_registry
from .resolvers import AssetResolver, AssetResolverSeries, AssetResolverVtk

__all__ = [
    "AssetKind",
    "AssetResolver",
    "AssetResolverRegistry",
    "AssetResolverSeries",
    "AssetResolverVtk",
    "Extra",
    "MetaAsset",
    "asset",
    "asset_resolver_registry",
    "get_assets",
    "get_inputs",
    "get_outputs",
    "get_temporaries",
    "input",
    "model_dump_without_assets",
    "output",
    "resolvers",
    "temporary",
]



import functools
from collections.abc import Callable
from typing import Any, Self, overload, override


class PluginCachedProperty[T](functools.cached_property[T]):
    @overload
    def __get__(self, instance: None, owner: type | None = None) -> Self: ...
    @overload
    def __get__(self, instance: object, owner: type | None = None) -> T: ...
    @override
    def __get__(self, instance: object | None, owner: type | None = None) -> Self | T:
        if instance is None:
            return super().__get__(instance, owner)
        while (parent := getattr(instance, "_plugin_parent", None)) is not None:
            instance = parent
        return super().__get__(instance, owner)

    @override
    def __set__(self, instance: object, value: T) -> None:
        assert self.attrname is not None
        instance.__dict__[self.attrname] = value


class PluginProperty[T](property):
    @overload
    def __get__(self, instance: None, owner: type, /) -> Self: ...
    @overload
    def __get__(self, instance: Any, owner: type | None = None, /) -> Any: ...
    def __get__(
        self, instance: object | None, owner: type | None = None, /
    ) -> Self | T:
        if instance is None:
            return super().__get__(instance, owner)
        while (parent := getattr(instance, "_plugin_parent", None)) is not None:
            instance = parent
        return super().__get__(instance, owner)


def plugin_cached_property[T](func: Callable[[Any], T]) -> PluginCachedProperty[T]:
    return PluginCachedProperty(func)


def plugin_property[T](fget: Callable[[Any], T]) -> PluginProperty[T]:
    return PluginProperty(fget)



from collections.abc import Generator
from pathlib import Path
from typing import override

from ._abc import AssetResolver


class AssetResolverSeries(AssetResolver):
    @override
    def match(self, path: Path) -> bool:
        return path.suffix == ".series"

    @override
    def resolve(self, path: Path) -> Generator[Path]:
        if (folder := path.with_suffix(".d")).exists():
            yield folder
        if (folder := path.with_suffix("")).exists():
            yield folder
